@@ -4,7 +4,7 @@
 @section LICENSE
 
 Greenflare SEO Web Crawler (https://greenflare.io)
-Copyright (C) 2020  Benjamin Görler. This file is part of
+Copyright (C) 2020-2021 Benjamin Görler. This file is part of
 Greenflare, an open-source project dedicated to delivering
 high quality SEO insights and analysis solutions to the world.
 
@@ -23,9 +23,12 @@ along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """
 
 from lxml.html import fromstring
-from .gflarerobots import GFlareRobots
+from greenflare.core.gflarerobots import GFlareRobots
 from requests import status_codes
-import urllib.parse
+from requests.utils import requote_uri
+from requests.compat import urlunparse
+from urllib.parse import urljoin
+from urllib3.util import parse_url
 from re import match, escape
 from functools import wraps
 from time import time
@@ -37,30 +40,31 @@ class GFlareResponse:
         self.settings = settings
         self.all_items = columns
         self.response = None
-        self.url = None
-        self.url_components = None
+        self.url = ''
+        self.base_url = ''
+        self.host = ''
         self.robots_txt_ua = "Googlebot"
         self.gfrobots = GFlareRobots('', self.settings.get("USER_AGENT", ''))
-        self.robots_txt_status = None
 
         self.spider_links = "Spider" in self.settings.get("MODE", "")
-        if self.robots_txt_status == "BLOCKED":
-            self.spider_links = False
-
+        
         self.extraction_separator = self.settings.get(
             'EXTRACTION_SEPARATOR', '; ')
 
         self.xpath_mapping = {
+            'href_all': '//a/@href',
+            'href_respect_nofollow': '//a[not(contains(@rel, "nofollow"))]/@href',
             'canonical_tag': '/html/head/link[@rel="canonical"]/@href',
-            'hreflang': '/html/head/link[@rel="alternate"]/@href',
-            'pagination': '/html/head/link[@rel="next"]/@href|//link[@rel="prev"]/@href',
+            'hreflang': '/html/head/link[@rel="alternate" and @hreflang]/@href',
+            'pagination': '/html/head/link[@rel="next"]/@href|/html/head/link[@rel="prev"]/@href',
             'images': '//img/@src',
             'stylesheets': '//link[@rel="stylesheet"]/@href',
             'javascript': '//script/@src',
             'h1': '//h1/text()',
             'h2': '//h2/text()',
             'page_title': '/html/head/title/text()',
-            'meta_description': '/html/head/meta[@name="description"]/@content'
+            'meta_description': '/html/head/meta[@name="description"]/@content',
+            'base_url': '/html/head/base/@href'
         }
 
         self.xpath_link_extraction = self.get_link_extraction_xpath()
@@ -68,6 +72,8 @@ class GFlareResponse:
         self.exclusions_regex = self.exclusions_to_regex(
             self.settings.get('EXCLUSIONS', []))
 
+        self.crawlable_schemes = ('http', 'https', '')
+
     def timing(f):
         @wraps(f)
         def wrap(*args, **kw):
@@ -80,12 +86,12 @@ class GFlareResponse:
 
     def set_response(self, response):
         self.response = response
-        # requests.get() encodes spaces within the path with %25
-        # If we encode the path beforehand, request.get() will double encode the path again resulting in the generation of endless new urls
-        # we need to decode the path back to what it was before request.get()
-        # encoded it
-        self.url = self.url_components_to_str(
-            self.parse_url(self.unencode_url(self.response.url)))
+        self.url = self.response.url
+        self.host = self.get_domain(self.url)
+        self.spider_links = "Spider" in self.settings.get("MODE", "")
+
+        if self.get_robots_txt_status == 'blocked' and 'respect_robots_txt' in self.settings.get('CRAWL_ITEMS', ''):
+            self.spider_links = False
 
         if self.is_robots_txt():
             self.response_to_robots_txt()
@@ -105,11 +111,14 @@ class GFlareResponse:
 
     def get_link_extraction_xpath(self):
 
-        xpaths = []
-        xpaths.append('//a/@href')
+        xpaths = []        
 
         crawl_items = self.settings['CRAWL_ITEMS']
-
+        
+        if not 'respect_nofollow' in crawl_items:
+            xpaths.append(self.xpath_mapping['href_all'])
+        else:
+            xpaths.append(self.xpath_mapping['href_respect_nofollow'])    
         if 'canonical_tag' in crawl_items:
             xpaths.append(self.xpath_mapping['canonical_tag'])
         if 'hreflang' in crawl_items:
@@ -125,15 +134,15 @@ class GFlareResponse:
 
         return '|'.join(xpaths)
 
-    # @timing
     def get_data(self):
 
-        self.url_components = urllib.parse.urlsplit(self.url)
         d = {'url': self.url}
         d['data'] = self.get_header_info()
 
         if len(self.response.content) > 0:
             self.tree = self.get_tree()
+            self.base_url = self.get_base_url()
+            
             if self.spider_links:
                 d['links'] = self.extract_links()
             d['data'] = {**d['data'], **self.get_crawl_data()}
@@ -147,74 +156,58 @@ class GFlareResponse:
 
         return d
 
-    # @timing
     def get_tree(self):
         try:
             # We need to use page.content rather than page.text because
             # html.fromstring implicitly expects bytes as input.
             return fromstring(self.response.content)
         except Exception as e:
-            print("Error parsing", self.url, "with lxml")
+            print('Error parsing', self.url, 'with lxml')
             print(e)
 
-    def parse_url(self, url):
-        try:
-            scheme, netloc, path, query, frag = urllib.parse.urlsplit(
-                url.strip())
-        except:
-            print(f'Error parsing {url}')
-            return {"scheme": '', "netloc": '', "path": '', "query": '', "frag": ''}
-        if not scheme and not netloc:
-            # Hack needed as non RFC tel references are not detected by
-            # urlsplit
-            if path.startswith("tel:"):
-                path.replace("tel:", "")
-                scheme = "tel"
-            else:
-                absolute_url = urllib.parse.urljoin(self.url, url)
-                scheme, netloc, path, query, frag = urllib.parse.urlsplit(
-                    absolute_url)
-        if ':' in netloc:
-            if scheme == 'https' and ':443' in netloc:
-                netloc = netloc.replace(':443', '')
-            elif scheme == 'http' and ':80' in netloc:
-                netloc = netloc.replace(':80', '')
-
-        return {"scheme": scheme, "netloc": netloc, "path": path.strip(), "query": query, "frag": frag}
-
-    def url_components_to_str(self, comp):
-        url = str(urllib.parse.urlunsplit(
-            (comp["scheme"], comp["netloc"], comp["path"], comp["query"], "")))
-        if comp['path'] == '':
-            url += '/'
-        return url
-
-    def unencode_url(self, url):
-        parsed = self.parse_url(url)
-        parsed["path"] = urllib.parse.unquote(parsed["path"])
-        return self.url_components_to_str(parsed)
-
     def get_domain(self, url):
-        domain = self.parse_url(url)["netloc"]
-        if "www." in domain:
-            return domain.replace("www.", "")
+        try:
+            _, _, domain, _, _, _, _ = parse_url(url)
+        except:
+            return ''
+
+        if not domain:
+            return ''
+        
+        if domain.startswith('www.'):
+            return domain.replace('www.', '')
+        
         return domain
 
     def get_robots_txt_url(self, url):
-        comps = self.parse_url(url)
-        comps["path"] = "robots.txt"
-        return self.url_components_to_str(comps)
+        comps = parse_url(url)
+        url = requote_uri(urlunparse([comps.scheme, comps.host, 'robots.txt', None, comps.query, comps.fragment]))
+        return url
 
     def is_external(self, url):
-        if self.settings.get("ROOT_DOMAIN", "") == "":
+        if self.settings.get('MODE') == 'List':
             return False
-        return self.get_domain(url) != self.settings.get("ROOT_DOMAIN", "")
+
+        domain = self.get_domain(url)
+        
+        if not domain:
+            return False
+        
+        return domain != self.settings.get('ROOT_DOMAIN', '')
 
     def is_excluded(self, url):
         if self.exclusions_regex:
             return bool(match(self.exclusions_regex, url))
         return False
 
+    def get_base_url(self) -> str:
+        extraction = self.extract_xpath(self.xpath_mapping['base_url'])
+        
+        if extraction:
+            return self.sanitise_url(extraction[0], base_url=self.url)
+
+        return self.url
+
     def exclusions_to_regex(self, exclusions):
 
         rules = []
@@ -245,10 +238,10 @@ class GFlareResponse:
 
         if self.is_external(url):
             return False
-        return self.parse_url(url)["path"] == "/robots.txt"
+        return parse_url(url).path == '/robots.txt'
 
     def get_final_url(self):
-        return self.url_components_to_str(self.parse_url(self.response.url))
+        return self.response.url
 
     def get_text(self):
         return self.response.text
@@ -270,18 +263,23 @@ class GFlareResponse:
         }
         return header
 
-    def valid_url(self, components):
-        if not "http" in components['scheme']:
+    def valid_url(self, url):
+        try:
+            cmps = parse_url(url)
+            if cmps.scheme and not cmps.host:
+                return False
+            if not cmps.scheme and cmps.host:
+                return False
+        except:
             return False
 
-        url = self.url_components_to_str(components)
-
-        if ' ' in url:
+        if cmps.scheme and not cmps.scheme in self.crawlable_schemes:
             return False
 
         # Filter out external links if needed
-        if "external_links" not in self.settings.get("CRAWL_ITEMS", "") and self.is_external(url):
-            return False
+        if self.settings.get('MODE') != 'List':
+            if "external_links" not in self.settings.get("CRAWL_ITEMS", "") and self.is_external(url):
+                return False
 
         if self.is_excluded(url):
             return False
@@ -291,13 +289,53 @@ class GFlareResponse:
             return False
         return True
 
-    # @timing
+    def sanitise_url(self, url: str, base_url='') -> str:
+        """Cleans a given input URL and returns a RFC compliant URL as a string."""
+
+        if isinstance(url, bytes):
+            url = url.decode('utf8')
+        else:
+            url = str(url)
+
+        # Remove leading whitespaces from url
+        url = url.lstrip()
+        
+        if base_url:
+            url = urljoin(base_url, url)
+
+        try:
+            scheme, auth, host, port, path, query, fragment = parse_url(url)
+        except:
+            return None
+
+        # Carefully reconstruct the network location
+        netloc = auth or ''
+        if netloc:
+            netloc += '@'
+        netloc += host
+        if port:
+            # Only report on ports if they are used in a non-standard way
+            if scheme == 'http' and port == 80:
+                pass
+            elif scheme == 'https' and port == 443:
+                pass
+            else:
+                netloc += ':' + str(port)
+
+        # Bare domains aren't valid URLs.
+        if not path:
+            path = '/'
+
+        url = requote_uri(urlunparse([scheme, netloc, path, None, query, fragment]))
+
+        # Search engines ignore hash fragments hence we remove them from URLs
+        url = url.split('#')[0]
+
+        return url
+
     def extract_links(self):
-        parsed_links = [self.parse_url(l) for l in self.extract_xpath(
-            self.xpath_link_extraction)]
-        links = list(set([self.url_components_to_str(l)
-                          for l in parsed_links if self.valid_url(l)]))
-        return links
+        links = [self.sanitise_url(url, base_url=self.base_url) for url in self.extract_xpath(self.xpath_link_extraction) if self.valid_url(url)]
+        return list(set(links))
 
     def get_txt_by_selector(self, selector, method="css", get="txt"):
         try:
@@ -353,7 +391,7 @@ class GFlareResponse:
             canonicals = self.extract_xpath(
                 self.xpath_mapping['canonical_tag'])
             if len(canonicals) > 0:
-                d['canonical_tag'] = canonicals[0]
+                d['canonical_tag'] = self.sanitise_url(canonicals[0], base_url=self.base_url)
             else:
                 d['canonical_tag'] = ''
 
@@ -395,7 +433,7 @@ class GFlareResponse:
     def is_canonicalised(self, url, canonical):
         if not canonical:
             return False
-        if self.url_components_to_str(self.parse_url(canonical)) != self.url_components_to_str(self.parse_url(url)):
+        if canonical != url:
             return True
         return False
 
@@ -461,38 +499,35 @@ class GFlareResponse:
     def has_redirected(self):
         return len(self.response.history) > 0
 
-    # @timing
     def get_redirects(self):
         data = []
         hist = self.response.history
 
-        if len(hist) > 0:
-            for i in range(len(hist)):
-                hob_url = self.url_components_to_str(
-                    self.parse_url(hist[i].url))
 
-                if 'external_links' not in self.settings.get('CRAWL_ITEMS', ''):
-                    if self.is_external(hob_url):
-                        break
+        for i in range(len(hist)):
+            hob_url = self.sanitise_url(hist[i].url)
 
-                robots_status = self.get_robots_txt_status(hob_url)
-                if 'respect_robots_txt' in self.settings.get('CRAWL_ITEMS', '') and 'follow_blocked_redirects' not in self.settings.get('CRAWL_ITEMS', '') and robots_status == 'blocked':
-                    continue
+            if 'external_links' not in self.settings.get('CRAWL_ITEMS', ''):
+                if self.is_external(hob_url):
+                    break
 
-                if i + 1 < len(hist):
-                    redirect_to_url = self.url_components_to_str(
-                        self.parse_url(str(hist[i + 1].url).strip()))
-                else:
-                    redirect_to_url = self.get_final_url()
+            robots_status = self.get_robots_txt_status(hob_url)
+            if 'respect_robots_txt' in self.settings.get('CRAWL_ITEMS', '') and 'follow_blocked_redirects' not in self.settings.get('CRAWL_ITEMS', '') and robots_status == 'blocked':
+                continue
 
-                hob_data = {"url": hob_url, "content_type": hist[i].headers.get('Content-Type', ""), 'status_code': hist[i].status_code, 'x_robots_tag': hist[
-                    i].headers.get('X-Robots-Tag', ''), 'redirect_url': redirect_to_url, 'robots_txt': robots_status}
+            if i + 1 < len(hist):
+                redirect_to_url = self.sanitise_url(str(hist[i + 1].url))
+            else:
+                redirect_to_url = self.get_final_url()
 
-                hob_data['crawl_status'] = self.get_full_status(
-                    hob_url, hob_data)
-                hob_row = self.dict_to_row(hob_data)
+            hob_data = {"url": hob_url, "content_type": hist[i].headers.get('Content-Type', ""), 'status_code': hist[i].status_code, 'x_robots_tag': hist[
+                i].headers.get('X-Robots-Tag', ''), 'redirect_url': redirect_to_url, 'robots_txt': robots_status}
 
-                data.append(hob_row)
+            hob_data['crawl_status'] = self.get_full_status(
+                hob_url, hob_data)
+            hob_row = self.dict_to_row(hob_data)
+
+            data.append(hob_row)
 
         return data
 
@@ -504,12 +539,6 @@ class GFlareResponse:
             return "allowed"
         return "blocked"
 
-    def attrib_to_list(self, xpath, attrib):
-        try:
-            return [self.url_components_to_str(self.parse_url(l.attrib[attrib])) for l in self.tree.xpath(xpath)]
-        except:
-            return []
-
     def extract_xpath(self, path):
         try:
             return self.tree.xpath(path)
