@@ -4,7 +4,7 @@
 @section LICENSE
 
 Greenflare SEO Web Crawler (https://greenflare.io)
-Copyright (C) 2020  Benjamin Görler. This file is part of
+Copyright (C) 2020-2021 Benjamin Görler. This file is part of
 Greenflare, an open-source project dedicated to delivering
 high quality SEO insights and analysis solutions to the world.
 
@@ -27,12 +27,12 @@ from greenflare.widgets.progresswindow import ProgressWindow
 from greenflare.widgets.filterwindow import FilterWindow
 from greenflare.widgets.enhancedentry import EnhancedEntry
 from greenflare.widgets.viewinlinks import ViewInlinks
-from greenflare.widgets.menuhelper import generate_menu
+from greenflare.widgets.helpers import generate_menu
 from greenflare.core.defaults import Defaults
-from concurrent import futures
-import functools
+from greenflare.widgets.helpers import run_in_background_with_window, tk_after
 from os import path, remove
 from webbrowser import open as open_in_browser
+from urllib.parse import urlsplit, urlunsplit
 import queue
 import sys
 
@@ -47,9 +47,6 @@ class CrawlTab(ttk.Frame):
         self.freeze_tabs = freeze_tabs
         self.unfreeze_tabs = unfreeze_tabs
 
-        self.executor = futures.ThreadPoolExecutor(max_workers=1)
-        self.win_progress = None
-
         self.topframe = ttk.Frame(self)
         self.topframe.pack(anchor='center', padx=20, pady=20, fill="x")
 
@@ -84,15 +81,13 @@ class CrawlTab(ttk.Frame):
         self.middle_frame.pack(anchor='center', fill='both', expand=True)
 
         self.treeview_table = ttk.Treeview(
-            self.middle_frame, selectmode="browse")
+            self.middle_frame, selectmode='browse')
 
-        # Capture right clicks on table
-        right_click = '<Button-3>'
+        # Capture left clicks on table
+        # Left clicks work best on all three platforms (right clicks are tricky on darwin)
+        left_click = '<Button-1>'
 
-        if sys.platform == 'darwin':
-            right_click = '<Button-2>'
-
-        self.treeview_table.bind(right_click, self.assign_treeview_click)
+        self.treeview_table.bind(left_click, self.assign_treeview_click)
 
         self.scrollbar_vertical = ttk.Scrollbar(
             self.middle_frame, orient="vertical", command=self.vertical_scrollbar_clicked)
@@ -114,11 +109,13 @@ class CrawlTab(ttk.Frame):
         self.label_urls_per_second.pack(side=LEFT)
 
         self.urls_crawled_string_var = StringVar()
-        self.urls_crawled_string_var.set("URLs crawled/discovered: 0/0")
+        self.urls_crawled_string_var.set("URLs crawled/discovered: 0/0 (0%)")
         self.label_urls_crawled = ttk.Label(
             self.bottom_frame, textvariable=self.urls_crawled_string_var)
         self.label_urls_crawled.pack(side=RIGHT)
 
+        self.percentage = 0
+
         self.populate_columns()
         self.row_counter = 1
 
@@ -143,28 +140,6 @@ class CrawlTab(ttk.Frame):
         self.viewed_table = 'crawl'
         self.filters = []
 
-    def daemonize(title=None, msg=None):
-        def decorator(target):
-            @functools.wraps(target)
-            def wrapper(*args, **kwargs):
-                args[0].win_progress = ProgressWindow(title=title, msg=msg)
-                args[0].win_progress.focus_force()
-                args[0].win_progress.grab_set()
-                result = args[0].executor.submit(target, *args, **kwargs)
-                result.add_done_callback(args[0].daemon_call_back)
-                return result
-
-            return wrapper
-
-        return decorator
-
-    def daemon_call_back(self, future):
-        self.win_progress.grab_release()
-        self.win_progress.destroy()
-        exception = future.exception()
-        if exception:
-            raise exception
-
     def clear_output_table(self):
         # Clear table
         self.treeview_table.delete(*self.treeview_table.get_children())
@@ -234,33 +209,34 @@ class CrawlTab(ttk.Frame):
             self.after(10, self.add_to_outputtable)
             self.after(10, self.update_buttons)
 
-    @daemonize(title="Stopping crawl ...", msg="Waiting for crawl to finish ...")
+    @run_in_background_with_window([], title='Stopping crawl ...', msg='Waiting for crawl to finish ...')
     def stop_crawl(self):
         self.crawler.end_crawl_gracefully()
         self.after(10, self.update_buttons)
 
     def btn_crawl_pushed(self):
-        url = self.entry_url_input.entry.get().strip()
-        if self.button_crawl["text"] == "Start":
+        url = self.entry_url_input.get().strip()
+
+        if self.button_crawl['text'] == 'Start':
             # Validate input url
-            url_components = self.crawler.gf.parse_url(url)
+            url_components = urlsplit(url)
 
             if self.crawler.settings.get('MODE', '') == 'Spider':
 
-                if url_components['scheme'] == '':
+                if url_components.scheme == '':
                     url = 'http://' + url
-                    url_components = self.crawler.gf.parse_url(url)
-                    print(url_components, url)
+                    url_components = urlsplit(url)
 
-                if url_components['netloc'] == '' or ' ' in url_components['netloc']:
+                if url_components.netloc == '' or ' ' in url_components.netloc:
                     messagebox.showerror(
                         title='Invalid URL', message='Please enter a valid URL!')
                     return
 
-            url = self.crawler.gf.url_components_to_str(url_components)
+                url = urlunsplit(url_components)
+                url = self.crawler.gf.sanitise_url(url, base_url='')
 
-            self.entry_url_input.entry.delete(0, 'end')
-            self.entry_url_input.entry.insert(0, url)
+                self.entry_url_input.entry.delete(0, 'end')
+                self.entry_url_input.entry.insert(0, url)
 
             self.start_new_crawl(url)
 
@@ -270,14 +246,13 @@ class CrawlTab(ttk.Frame):
         elif self.button_crawl["text"] == "Resume":
             self.populate_columns()
             self.crawler.resume_crawl()
+            self.row_counter = self.crawler.urls_crawled + 1
             self.after(10, self.add_to_outputtable)
             self.after(10, self.update_buttons)
 
         elif self.button_crawl["text"] == "Restart":
             self.start_new_crawl(url)
 
-        print(self.crawler.settings)
-
     def update_buttons(self):
         btn_txt = self.button_crawl["text"]
         if btn_txt == "Start":
@@ -300,17 +275,10 @@ class CrawlTab(ttk.Frame):
         msg = messagebox.askquestion ('Clear View','Are you sure you want clear the view? All data has been saved.', icon = 'warning')
         
         if msg == 'yes':
-            self.progressbar["value"] = 0
-            self.clear_output_table()
-            self.populate_columns()
-            self.button_crawl["text"] = "Start"
-            self.suspend_auto_scroll = False
-            self.filters = None
-            self.reset_filter_window()
-            self.entry_url_input.entry['state'] = 'enabled'
+            self.reset(reset_url_input=False)
         else:
-            pass
-
+            return
+        
     def add_item_to_outputtable(self, item):
         self.treeview_table.insert(
             '', 'end', text=self.row_counter, values=item)
@@ -332,6 +300,7 @@ class CrawlTab(ttk.Frame):
 
         if len(items) > 0:
             for item in items:
+                # item = self.remove_unicode_from_item(item)
                 self.add_item_to_outputtable(item)
 
             self.update_progressbar()
@@ -352,6 +321,7 @@ class CrawlTab(ttk.Frame):
             messagebox.showerror(title='Error - Timed Out',
                                  message='Crawl timed out!')
             self.button_crawl["text"] = "Resume"
+            self.button_clear['state'] = 'enabled'
             self.update_bottom_stats()
             return
         if self.crawler.crawl_running.is_set():
@@ -360,10 +330,14 @@ class CrawlTab(ttk.Frame):
 
         self.after(250, self.add_to_outputtable)
 
-    # @daemonize(title="Loading crawl ...", msg="Please wait while the crawl is loading ...")
+    @tk_after
+    def add_items(self, items):
+        for i, item in enumerate(items, 1):
+            self.treeview_table.insert('', 'end', text=i, values=item)
+
+    @run_in_background_with_window([], title='Loading crawl ...', msg='Please wait while the data is being loaded ...')
     def load_crawl_to_outputtable(self, filters, table, columns=None):
         self.suspend_auto_scroll = True
-        print('load_crawl_to_outputtable', filters, table, columns)
 
         if not filters:
             self.master.title(
@@ -377,8 +351,7 @@ class CrawlTab(ttk.Frame):
             filters, table, columns)
         self.populate_columns(columns=columns)
 
-        for item in items:
-            self.add_item_to_outputtable(item)
+        self.add_items(items)
 
         self.suspend_auto_scroll = False
         self.treeview_table.yview_moveto(1)
@@ -386,16 +359,16 @@ class CrawlTab(ttk.Frame):
     def update_progressbar(self):
         with self.lock:
             if self.crawler.urls_total > 0:
-                percentage = int((self.crawler.urls_crawled /
+                self.percentage = int((self.crawler.urls_crawled /
                                   self.crawler.urls_total) * 100)
-                self.progressbar["value"] = percentage
+                self.progressbar["value"] = self.percentage
                 if sys.platform == "win32":
-                    self.style.configure('text.Horizontal.TProgressbar', text=f'{percentage} %')
+                    self.style.configure('text.Horizontal.TProgressbar', text=f'{self.percentage} %')
 
     def update_bottom_stats(self):
         with self.lock:
             self.urls_string_var.set(f"Speed: {self.crawler.current_urls_per_second} URL/s")
-            self.urls_crawled_string_var.set(f"URLs crawled/discovered: {self.crawler.urls_crawled}/{self.crawler.urls_total}")
+            self.urls_crawled_string_var.set(f"URLs crawled/discovered: {'{:,}'.format(self.crawler.urls_crawled)}/{'{:,}'.format(self.crawler.urls_total)} ({self.percentage}%)")
 
     def update(self):
         self.button_crawl["text"] = "Resume"
@@ -407,10 +380,13 @@ class CrawlTab(ttk.Frame):
         self.populate_columns()
         self.update_progressbar()
 
-    def reset(self):
+    def reset(self, reset_url_input=True):
         self.entry_url_input.entry["state"] = "enabled"
-        self.entry_url_input.entry.delete(0, 'end')
-        self.entry_url_input.entry.insert(0, "Enter URL to crawl")
+        
+        if reset_url_input:
+            self.entry_url_input.entry.delete(0, 'end')
+            self.entry_url_input.entry.insert(0, "Enter URL to crawl")
+        
         self.progressbar["value"] = 0
         self.clear_output_table()
         self.populate_columns()
@@ -418,6 +394,7 @@ class CrawlTab(ttk.Frame):
         self.suspend_auto_scroll = False
         self.filters = None
         self.reset_filter_window()
+        self.crawler.reset_crawl()
         self.master.title(Defaults.window_title)
 
     def show_list_mode(self):
@@ -453,10 +430,12 @@ class CrawlTab(ttk.Frame):
             col = self.treeview_table.identify_column(event.x)
             self.selected_column = self.treeview_table.heading(col)['text']
 
-            try:
-                self.popup_menu.tk_popup(event.x_root, event.y_root + 20, 0)
-            finally:
-                self.popup_menu.grab_release()
+            # Only show a context menu if the table is not empty
+            if len(self.treeview_table.get_children()) > 0:
+                try:
+                    self.popup_menu.tk_popup(event.x_root, event.y_root + 20, 0)
+                finally:
+                    self.popup_menu.grab_release()
 
     def show_filter_window(self, label):
         columns = self.get_display_columns()
